#include <stdio.h>
#include <stdlib.h>
#include "procs.h"

#define DEBUGGING_FREAD 0

/* This function reads size * count characters into buffer,        *
 * then swaps the byte order for each object read,                 *
 * in order to swap byteorder between the data and the machine.	   *
 * That is, this function swaps bytes to go from least-significant *
 * byte first to most significant byte first.                      */
size_t fread_swab(void *buffer, size_t size, size_t count, FILE *input)
{
  int  max_to_read_at_once = 1000;
  char temp         = 0;
  int  objects_read = 0;
  int  i            = 0;
  int  first        = 0;
  int  last         = 0;
  int  numToRead    = 0;
  int  numRead      = 0;

  objects_read = fread( buffer, size, count, input );

#if 0
  /*if (count == 1) {*/
  if (count <= max_to_read_at_once) {
    objects_read = fread( buffer, size, count, input );
  }
  else {
    switch (size) {
    case 1: /* Cast buffer to char * */
      while (i < count) {
        /* Read up to max_to_read_at_once characters at a time */
        numToRead = ((count - i) > max_to_read_at_once) ?
                    max_to_read_at_once :
                    count - i;
        numRead = fread( (void *)( &(((char *)buffer)[i]) ), size, numToRead, input );
        if (numRead != numToRead) {
          fprintf(stderr,
                  "Error w/ SEG2 File %s: tried to read %d objs of %d bytes, but only read %d.\n",
                  *input, numToRead, size, numRead);
        }
#ifdef DEBUGGING_FREAD
        fprintf(stderr, "1 byte buffer[%d] = 0x%X\n", i, ((char *)buffer)[i]);
#endif
        objects_read += numRead;
        i            += numRead;
      }
      break;
    case 2: /* Cast buffer to short int * */
      while (i < count) {
        /* Read up to max_to_read_at_once characters at a time */
        numToRead = ((count - i) > max_to_read_at_once) ?
                    max_to_read_at_once :
                    count - i;
        numRead = fread( (void *)( &(((short int *)buffer)[i]) ), size, numToRead, input );
        if (numRead != numToRead) {
          fprintf(stderr,
                  "Error w/ SEG2 File %s: tried to read %d objs of %d bytes, but only read %d.\n",
                  *input, numToRead, size, numRead);
        }
#ifdef DEBUGGING_FREAD
        fprintf(stderr, "2 byte buffer[%d] = 0x%X\n", i, ((short int *)buffer)[i]);
#endif
        objects_read += numRead;
        i            += numRead;
      }
      break;
    case 4: /* Cast buffer to int * */
      while (i < count) {
        /* Read up to max_to_read_at_once characters at a time */
        numToRead = ((count - i) > max_to_read_at_once) ?
                    max_to_read_at_once :
                    count - i;
        numRead = fread( (void *)( &(((int *)buffer)[i]) ), size, numToRead, input );
        if (numRead != numToRead) {
          fprintf(stderr,
                  "Error w/ SEG2 File %s: tried to read %d objs of %d bytes, but only read %d.\n",
                  *input, numToRead, size, numRead);
        }
#ifdef DEBUGGING_FREAD
        fprintf(stderr, "4 byte buffer[%d] = 0x%X\n", i, ((int *)buffer)[i]);
#endif
        objects_read += numRead;
        i            += numRead;
      }
      break;
    default: /* irregular size: just read the whole thing, as if count were less than max */
      objects_read = fread( buffer, size, count, input );
      break;
    }
  }
#endif

  /* Reverse the byte order of each element of the read data */
  if (size > 1) {
    for (i = 0; i < objects_read; i++) {
      for ( first = i * size, last = first + size - 1;
            first < last;
            first++, last-- ) {
        temp = ((char *)buffer)[first];
        ((char *)buffer)[first] = ((char *)buffer)[last];
        ((char *)buffer)[last]  = temp;
      }
    }
  }

  return(objects_read); /* should be less than or equal to count */
}
