#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <string.h>
#include "seg2_struct.h"
#include "externals.h"
#include "procs.h"



KEY_VAL_PAIR * get_key_val_pair(SEG2FILEDESC *file_desc,
                                char         *strings,
                                int           index,
                                int          *chars_to_next_pair)
{
  KEY_VAL_PAIR *key_val = NULL;
  short int block_ID    = file_desc->block_ID;
  char     *strTerm     = file_desc->strTerm;
  char      strTermSize = file_desc->strTermSize;
  short int offset;
  int       index_to_next;
  int       max_str_length;
  char     *key;
  char     *val;
  int       key_length = 0;
  int       val_length = 0;
  char      key_name_violation = 0; /* Set to 1 if the keyword has any lowercase letters */

  memcpy (&offset,&strings[index],2);
  if (block_ID == OPPOSITE_WORDORDER_ID) swab2 (&offset);
  *chars_to_next_pair = (int)offset;
  if (offset == 0) { /* This is the last string and it is empty.  Return immediately. */
    return key_val;
  }

  index_to_next = index + offset;
  max_str_length = offset - 2 - strTermSize; /* (2 == # of chars in offset) */
  key = (char *)malloc(sizeof(char) * max_str_length); /* Max possible array length, until actual length is known. */
  val = (char *)malloc(sizeof(char) * max_str_length); /* Max possible array length, until actual length is known. */

  index += 2; /* skip past the offset, which was just read */

  /* Read the keyword until the space or tab (which seperates keyword from value) is found */
  while ( (strings[index] != ' ' ) &&
          (strings[index] != '\t') &&
          (! end_of_string(strings, index, strTerm, strTermSize)) &&
          (index < index_to_next) ) {
    if (islower(strings[index]))  key_name_violation = 1;
    key[key_length++] = strings[index++];
  }

  if (key_name_violation) {
    /* This is just a warning, but won't prevent processing of the data */
    fprintf(stderr,
            "SEG-2 strings format WARNING: keyword (%s) should not contain lowercase (but will still be processed).\n",
            key);
  }

  /* Skip the blanks and tabs until the value starts */
  while ( ((strings[index] == ' ' ) ||
           (strings[index] == '\t')  ) &&
          (! end_of_string(strings, index, strTerm, strTermSize)) &&
          (index < index_to_next) ) {
    index++;
  }

  /* Get the value until the string terminator or the index reaches the next string */
  while ( (! end_of_string(strings, index, strTerm, strTermSize)) &&
          (index < index_to_next) ) {
    val[val_length++] = strings[index++];
  }

  /* Allocate memory for the KEY_VAL_PAIR structure */
  key_val = (KEY_VAL_PAIR *)malloc(sizeof(KEY_VAL_PAIR));
  if (key_val == NULL) {
    fprintf(stderr,
            "get_key_val_pair: Can't allocate memory block for key_val structure. current keyword: %s\n",
            key);
    return(NULL);
  } 
  /* Allocate memory for the KEY_VAL_PAIR structure's keyword array */
  key_val->key = (char *)malloc(sizeof(char) * (key_length + 1));
  if (key_val->key == NULL) {
    fprintf(stderr,
            "get_key_val_pair: Can't allocate memory block for key_val's key array. current keyword: %s\n",
            key);
    free(key_val);
    return(NULL);
  }
  /* Allocate memory for the KEY_VAL_PAIR structure's value array */
  key_val->val = (char *)malloc(sizeof(char) * (val_length + 1));
  if (key_val->val == NULL) {
    fprintf(stderr,
            "get_key_val_pair: Can't allocate memory block for key_val's val array. current keyword: %s\n",
            key);
    free(key_val->key);
    free(key_val);
    return(NULL);
  }

  /* Copy the keyword and value into the KEY_VAL_PAIR structure */
  strncpy(key_val->key, key, key_length);
  strncpy(key_val->val, val, val_length);
  key_val->key[key_length] = '\0';
  key_val->val[val_length] = '\0';
  key_val->key_length = key_length;
  key_val->val_length = val_length;

  free(key);
  free(val);

  return(key_val); 
}

char end_of_string(char * strings, int index, char * strTerm, char strTermSize) {
  char returnval = FALSE;

  if ( ((strTermSize == 1) && (strings[index] == strTerm[0])) ||
       ((strTermSize == 2) && (strings[index] == strTerm[0]) && (strings[index+1] == strTerm[1])) )
    returnval = TRUE;

  return returnval;
}

/*******************************************************************************
* FUNCTION: categorize_keys:                                                   *
* INPUTS:   KEY_VAL_PAIR   keys_and_vals[]                                     *
*           int            num_keys                                            *
*           int            key_categories[] (values are filled in by function) *
*           SEG2_KEYWORDS *keywords_list                                       *
* OUTPUTS:  int            key_categories[]                                     *
* RETURNS:  int            unknown_keys_count                                   *
* DESCRIPTION: The key of each KEY_VAL_PAIR in keys_and_vals[], is compared to *
* the key names in the keywords_list[].                                        *
*   If a match is found, then the index of that key in the keyword_list[] is   *
* stored in the key_categories[] array, at the same index as the KEY_VAL_PAIR  *
* being examined in the keys_and_vals[] array.                                 *
*   If a match is not found for the current key/val pair, then its             *
* corresponding entry in key_categories gets the value -1, indicating that the *
* keyword is unknown.                                                          *
*   When this function is done, it returns the number of unknown keys found.   *
*   The calling function will then use all this information to store the       *
* KEY_VALUE_PAIRs into the file descriptor or trace descriptor structure's     *
* parsed_strings known or unknown arrays.  The known arrays can be indexed     *
* with the constants corresponding to the known keywords, to quickly get at    *
* their value without having to search through the strings for a matching      *
* keyword, which would get inefficient if the same set of string is searched   *
* through multiple times.                                                      *
*******************************************************************************/
int categorize_keys (KEY_VAL_PAIR **keys_and_vals,
                     int            num_keys,
                     int           *key_categories,
                     SEG2_KEYWORDS *keywords_list) {
  int  key_val_index;
  int  known_keys_index;
  int  key_search_start_index;
  int  unknown_keys_count = 0; /* count of key/value pairs whose keys didn't match known keys */
  char search_for_key;         /* boolean TRUE or FALSE */
  int  comparison;

  /* First, clear out the category list, flagging all the keys as unknown */
  for (key_val_index = 0; key_val_index < num_keys; key_val_index++) {
    key_categories[key_val_index] = -1;
  }

  /* For each KEY_VAL_PAIR of keys_and_vals[] */
  for (key_val_index = 0; key_val_index < num_keys; key_val_index++) {
    /* set up keyword search parameters */
    search_for_key = TRUE;
    key_search_start_index = known_keys_index = keywords_list->current_index;
  
    /* try to match the key of the KEY_VAL_PAIR to the *
     * known keywords in keywords_list->keywords[].    */
    while (search_for_key == TRUE) {
      comparison = strcmp((keys_and_vals[key_val_index])->key, keywords_list->keywords[known_keys_index]);
      if (comparison == 0) { /* keyword is matched */
        key_categories[key_val_index] = known_keys_index;

        /* The list of known keywords is treated like a circular buffer, ensuring that all  *
         * known keywords present can be found, even if they are not stored alphabetically. */
        known_keys_index = (known_keys_index < keywords_list->max_index) ? known_keys_index + 1 : 0;

        /* The next keyword search will start at the word after the last one found. */
        keywords_list->current_index = known_keys_index;

        search_for_key = FALSE;
      }
      else { /* keyword is not yet matched.  technically they should be *
              * ordered alphabetically, but I won't be that strict.     */
        /* The list of known keywords is treated like a circular buffer, ensuring that all  *
         * known keywords present can be found, even if they are not stored alphabetically. */
        known_keys_index = (known_keys_index < keywords_list->max_index) ? known_keys_index + 1 : 0;

        if (known_keys_index == key_search_start_index) {
          /* All kown keywords have been searched and none matched this one. *
           * Add it to the list of unknown keys, all of which will be added  *
           * to file_desc at the end, when a total count is available.       */
          key_categories[key_val_index] = -1;
          unknown_keys_count++;
          search_for_key = FALSE;
        }
      }
    }
  }

  return (unknown_keys_count);
}
